import threading
from gai.lib.common import logging, generators_utils
import os
from dotenv import load_dotenv
load_dotenv()
logger = logging.getLogger(__name__)
from gai.ttt.server.gai_exllamav2 import GaiExLlamav2

class SingletonHost:
    __instance = None       # singleton

    @staticmethod
    def GetInstanceFromPath(generator_name,config_path=None):
        """Static method to access this singleton class's instance."""
        if SingletonHost.__instance == None:
            config_path=os.path.expanduser(config_path)
            config = generators_utils.load_generators_config(config_path)
            SingletonHost.__instance=SingletonHost(config[generator_name])
        return SingletonHost.__instance

    @staticmethod
    def GetInstanceFromConfig(config):
        """Static method to access this singleton class's instance."""
        if SingletonHost.__instance == None:
            SingletonHost.__instance=SingletonHost(config)
        return SingletonHost.__instance

    def __init__(self,config):
        """Virtually private constructor."""
        if SingletonHost.__instance is not None:
            raise Exception(
                "SingletonHost: This class is a singleton! Access using GetInstance().")
        else:
            # This class only has 4 attributes

            # config is always the first to be loaded from constructor
            self.config = config

            # generator is loaded by calling load()
            self.generator = None

            # generator_name matches config["generator_name"] and is loaded only when self.generator is successfully loaded 
            self.generator_name = None

            # Finally, for thread safety and since this is run locally, use semaphore to ensure only one thread can access the generator at a time
            self.semaphore = threading.Semaphore(1)

            SingletonHost.__instance = self

    def __enter__(self):
        self.load()
        return self
    
    def __exit__(self,exc_type, exc_value,traceback):
        self.unload()
        import gc,torch
        gc.collect()
        torch.cuda.empty_cache()

    # This is idempotent
    def load(self):

        if self.generator_name == self.config["generator_name"]:
            logger.debug(
                "SingletonHost.load: Generator is already loaded. Skip loading.")
            return self

        if self.generator_name and self.generator_name != self.config["generator_name"]:
            logger.debug(
                "SingletonHost.load: New generator_name specified, unload current generator.")
            if self.generator:
                self.unload()
            return self

        if self.config["generator_name"] == "exllamav2-mistral7b" and self.config["type"] == "ttt":
            try:
                target_name=self.config["generator_name"]
                logger.info(f"SingletonHost: Loading generator {target_name}...")
                self.generator = GaiExLlamav2(gai_config=self.config)
                self.generator.load()
                self.generator_name=target_name
                return self
            except Exception as e:
                self.unload()
                logger.error(
                    f"SingletonHost: Error loading generator {self.generator_name}: {e}")
                raise e
            
    def unload(self):
        if self.generator is not None:
            self.generator.unload()
            del self.generator
            self.generator = None
            self.generator_name = None
            import gc,torch
            gc.collect()
            torch.cuda.empty_cache()
        return self

    def create(self, **model_params):
        if self.generator is None:
            logger.error("SingletonHost.create: Generator is not loaded.")
            raise Exception("SingletonHost.create: Generator is not loaded.")
        with self.semaphore:
            return self.generator.create(**model_params)

    def token_count(self, text):
        if self.generator is None:
            logger.error("SingletonHost.create: Generator is not loaded.")
            raise Exception("SingletonHost.create: Generator is not loaded.")
        if hasattr(self.generator, 'token_count'):
            return self.generator.token_count(text)
        raise Exception("token_count is not supported by this generator.")

    def get_token_ids(self, text):
        if self.generator is None:
            logger.error("SingletonHost.create: Generator is not loaded.")
            raise Exception("SingletonHost.create: Generator is not loaded.")
        if hasattr(self.generator, 'get_token_ids'):
            return self.generator.get_token_ids(text)
        raise Exception("get_token_ids is not supported by this generator.")

    async def index_async(self, 
                          collection_name, 
                          file_path, 
                          file_type=None,
                          title='', 
                          source= '', 
                          abstract='',
                          authors='',
                          publisher ='',
                          published_date='', 
                          comments='',
                          keywords='',
                          chunk_size=None, 
                          chunk_overlap=None, 
                          status_publisher=None):
        if self.generator is None:
            logger.error("SingletonHost.create: Generator is not loaded.")
            raise Exception("SingletonHost.create: Generator is not loaded.")

        if self.generator.generator_name != "rag":
            logger.error(
                f"SingletonHost.index: The generator {self.generator.generator_name} does not support indexing.")
            raise Exception(
                f"SingletonHost.index: The generator {self.generator.generator_name} does not support indexing.")
        with self.semaphore:
            return await self.generator.index_async(
                collection_name=collection_name, 
                file_path=file_path,
                file_type=file_type,
                title=title,
                source=source,
                abstract=abstract,
                authors=authors,
                publisher=publisher,
                published_date=published_date,
                comments=comments,
                keywords=keywords,
                chunk_size=chunk_size,
                chunk_overlap=chunk_overlap,
                status_publisher=status_publisher)

    def retrieve(self, collection_name, query_texts, n_results=None):
        if self.generator is None:
            logger.error("SingletonHost.create: Generator is not loaded.")
            raise Exception("SingletonHost.create: Generator is not loaded.")

        if self.generator.generator_name != "rag":
            logger.error(
                f"SingletonHost.retrieve: The generator {self.generator.generator_name} does not support retrieval.")
            raise Exception(
                f"SingletonHost.retrieve: The generator {self.generator.generator_name} does not support retrieval.")
        with self.semaphore:
            return self.generator.retrieve(collection_name, query_texts, n_results)




from gai.ttt.server.builders import OutputMessageBuilder
from gai.ttt.server.builders import OutputChunkBuilder

class CompletionsFactory:

    def __init__(self):
        self.message = OutputMessageBuilder()
        self.chunk = OutputChunkBuilder()





import json
from openai.types.chat.chat_completion import ChatCompletion, ChatCompletionMessage, Choice , CompletionUsage
from openai.types.chat.chat_completion_message_tool_call import ChatCompletionMessageToolCall
from openai.types.chat.chat_completion_message_tool_call_param import Function
from datetime import datetime
from uuid import uuid4
import re
from jsonschema import validate, ValidationError

class OutputMessageBuilder:
    """
    # Documentation
    Descriptions: This class is used to build an OpenAI-styled ChatCompletion object to be returned from text generation.
    It is used to maintain compatibility with the OpenAI API design to facilitate drop-in replacements.
    Example: Used by generating text generation and text streaming output.
    """

    def build_toolcall(self,result):
        state = "function_name"
        eos_reason=result["eos_reason"]
        if eos_reason=="stop_string":
            eos_reason="stop"
        if eos_reason=="stop_token":
            eos_reason="stop"
        if eos_reason=="max_new_tokens":
            eos_reason="length"

        from gai.lib.common.generators_utils import get_tools_schema
        text = result["full_completion"]
        jsoned = json.loads(text)
        schema = get_tools_schema()        
        try:
            validate(instance={
                "type":"function","function":jsoned
            }, schema=schema)

            function_name = None
            if state == "function_name":
                text = result["full_completion"]
                text = re.sub(r'\s+', ' ', text)
                function_name_pattern = r'\"name\"\s*:\s*\"(.*?)\",'
                match = re.search(function_name_pattern, text, re.DOTALL)
                if match:
                    function_name=match.group(1)
                    state = "function_args"

            function_arguments = None
            if state == "function_args":
                text = result["full_completion"]
                text = re.sub(r'\s+', ' ', text)
                function_arguments_pattern = r'"(parameters|arguments)"\s*:\s*({.*?})'
                match = re.search(function_arguments_pattern, text, re.DOTALL)
                if match:
                    function_arguments=match.group(2)
                    # Remove escape sequence from keys
                    function_arguments = function_arguments.replace("\\","")
                    # If not parsable, return None and continue streaming.
                    try:
                        function_arguments = json.dumps(json.loads(function_arguments))
                    except Exception as e:
                        return None

                    return OutputMessageBuilder(
                        ).add_chat_completion(generator="exllamav2-mistral7b"
                            ).add_choice(finish_reason='tool_calls'
                                ).add_tool(
                                    function_name=function_name,
                                    function_arguments=function_arguments
                                    ).add_usage(
                                        prompt_tokens=result["prompt_tokens"],
                                        new_tokens=result["new_tokens"]
                                        ).build()
        except ValidationError as e:
            return



    def build_content(self,result):
        eos_reason=result["eos_reason"]
        if eos_reason=="stop_string":
            eos_reason="stop"
        if eos_reason=="stop_token":
            eos_reason="stop"
        if eos_reason=="max_new_tokens":
            eos_reason="length"
        return OutputMessageBuilder(
            ).add_chat_completion(generator="exllamav2-mistral7b"
                ).add_choice(finish_reason=eos_reason,logprobs=None
                    ).add_content(
                        content=result["full_completion"]
                        ).add_usage(
                            prompt_tokens=result["prompt_tokens"],
                            new_tokens=result["new_tokens"]
                            ).build()

    def generate_chatcompletion_id(self):
        return "chatcmpl-"+str(uuid4())

    def generate_creationtime(self):
        return int(datetime.now().timestamp())

    def generate_toolcall_id(self):
        return "call_"+str(uuid4())

    def add_chat_completion(self,generator):
        try:
            chatcompletion_id = self.generate_chatcompletion_id()
            created = self.generate_creationtime()
            self.result = ChatCompletion(
                id=chatcompletion_id,
                choices=[],
                created=created,
                model=generator,
                object='chat.completion',
                usage=None
            )
            return self
        except Exception as e:
            print("OutputMessageBuilder.add_chat_completion:",e)
            raise e

    def add_choice(self,finish_reason,logprobs=None):
        try:
            self.result.choices.append(Choice(
                finish_reason=finish_reason,
                index=0,
                message=ChatCompletionMessage(role='assistant',content=None, function_call=None, tool_calls=[]),
                logprobs=logprobs,
            ))
            return self
        except Exception as e:
            print("OutputMessageBuilder.add_choice:",e)
            raise e
        

    def add_tool(self,function_name,function_arguments):
        try:
            toolcall_id = self.generate_toolcall_id()
            self.result.choices[0].message.tool_calls.append(ChatCompletionMessageToolCall(
                id = toolcall_id,
                function = Function(
                    name=function_name,
                    arguments=function_arguments
                ),
                type='function'
            ))
            return self
        except Exception as e:
            print("OutputMessageBuilder.add_tool:",e)
            raise e

    def add_content(self,content):
        try:
            self.result.choices[0].message.content = content
            self.result.choices[0].message.tool_calls = None
            return self
        except Exception as e:
            print("OutputMessageBuilder.add_content:",e)
            raise e
    
    def add_usage(self, prompt_tokens, new_tokens):
        try:
            total_tokens = prompt_tokens + new_tokens
            self.result.usage = CompletionUsage(
                prompt_tokens=prompt_tokens,
                completion_tokens=new_tokens,
                total_tokens=total_tokens
            )
            return self
        except Exception as e:
            print("OutputMessageBuilder.add_usage:",e)
            raise e
    
    def build(self):
        try:
            return self.result.copy()
        except Exception as e:
            print("OutputMessageBuilder.build:",e)
            raise e



from openai.types.chat.chat_completion_chunk import ChatCompletionChunk, Choice as ChunkChoice, ChoiceDelta
from openai.types.chat.chat_completion_chunk import ChoiceDeltaToolCall, ChoiceDeltaToolCallFunction
from datetime import datetime
from uuid import uuid4

class OutputChunkBuilder:

    def build_content(self,result):

        if (result is None):
            return OutputChunkBuilder(
                ).add_chunk(generator="exllamav2-mistral7b"
                    ).add_chunk_choice_delta(finish_reason=None, role="assistant"
                        ).add_chunk_choice_delta_content(content=''
                            ).build()

        if (type(result) is str):
            return OutputChunkBuilder(
                ).add_chunk(generator="exllamav2-mistral7b"
                    ).add_chunk_choice_delta(finish_reason=None, role=None
                        ).add_chunk_choice_delta_content(content=result
                            ).build()

        if (type(result) is dict):
            eos_reason=result["eos_reason"]
            if eos_reason=="stop_string":
                eos_reason="stop"
            if eos_reason=="stop_token":
                eos_reason="stop"
            if eos_reason=="max_new_tokens":
                eos_reason="length"
            return OutputChunkBuilder(
                ).add_chunk(generator="exllamav2-mistral7b"
                    ).add_chunk_choice_delta(finish_reason=eos_reason, role=None
                        ).build()

    def build_stream(self,streaming_response):
        def streamer():
            head =  self.build_content(None)
            yield head
            for chunk in streaming_response:
                yield self.build_content(chunk)
        return (chunk for chunk in streamer())

    def __init__(self, result=None):
        self.result = None
        if result:
            self.result = result.copy()

    def copy(self):
        return OutputChunkBuilder(self.result)

    def generate_chatcompletion_id(self):
        return "chatcmpl-"+str(uuid4())

    def generate_creationtime(self):
        return int(datetime.now().timestamp())

    def generate_toolCall_id(self):
        return "call_"+str(uuid4())

    def add_chunk(self,generator):
        chatcompletion_id = self.generate_chatcompletion_id()
        created = self.generate_creationtime()
        self.result = ChatCompletionChunk(
            id=chatcompletion_id,
            choices=[],
            created=created,
            model=generator,
            object='chat.completion.chunk'
        )
        return self

    def add_chunk_choice_delta(self, role=None, finish_reason=None):
        self.result.choices.append(
            ChunkChoice(
                delta=ChoiceDelta(
                    content=None, 
                    role=role, 
                    tool_calls=None, 
                    function_call=None
                    ),
                index=0,
                logprobs=None,
                finish_reason=finish_reason,
            )            
        )
        return self

    def add_chunk_choice_delta_content(self, content):
        self.result.choices[0].delta.content=content
        return self

    def add_chunk_choice_delta_toolcall_name(self, name):
        self.result.choices[0].delta.tool_calls=[ChoiceDeltaToolCall(
            index=0,
            id=self.generate_toolCall_id(),
            function=ChoiceDeltaToolCallFunction(
                name=name,
                arguments=''
                ),
            type='function'
            )]
        return self

    def add_chunk_choice_delta_toolcall_arguments(self, arguments):
        self.result.choices[0].delta.tool_calls=[ChoiceDeltaToolCall(
            index=0,
            id=None,
            function=ChoiceDeltaToolCallFunction(
                name=None,
                arguments=arguments
                ),
            type='function'
            )]
        return self
    
    def build(self):
        return self.result.copy()




from setuptools import setup,find_packages
import os

# Is this used for publishing?
is_publishing=os.environ.get("TWINE_USERNAME",False)

# Find base directory with reference to setup.py
def get_here():
    return os.path.dirname(os.path.abspath(__file__))
HERE = get_here()
print(f"\033[31mHERE: {HERE}\033[0m")

# Read the version from the VERSION file
def get_version():
    version_file = os.path.join(HERE, 'VERSION')
    with open(version_file, 'r') as f:
        return f.read().strip()
VERSION = get_version()
print(f"\033[31mInstalling: {VERSION}\033[0m")

# Read the requirements from the requirements.txt file
DEFAULT_REQUIREMENTS_PATH=os.path.join(HERE,"src/gai/lib/requirements.txt")
def parse_requirements(filename):
    required=[]
    filepath = os.path.join(HERE, filename)
    if os.path.exists(filepath):
        with open(filepath) as f:
            required = f.read().splitlines()
        return required
DEFAULT_REQUIREMENTS = parse_requirements("./src/gai/lib/requirements.txt")

if not DEFAULT_REQUIREMENTS or len(DEFAULT_REQUIREMENTS)==0:
    raise Exception(f"No requirements found in {DEFAULT_REQUIREMENTS_PATH}")

# Install local wheel file
def local_pkg(name: str,file_path: str) -> str:
    """Returns a path to a local package."""
    file_path=os.path.join(os.getcwd(),"wheels",file_path)
    return f"{name} @ file://{file_path}"

# Merge two lists and remove duplicates
def merge_list(list1, list2):
    # Ensure both list1 and list2 are lists, default to empty list if None
    list1 = list1 if list1 is not None else []
    list2 = list2 if list2 is not None else []
    return list(set(list1) | set(list2))


# For default and cli use
cli_requirements=DEFAULT_REQUIREMENTS.copy()

# For development, debugging and testing on local machine
dev_requirements = merge_list(DEFAULT_REQUIREMENTS,[
    "pytest",
    "ipykernel",
    "pynvml",
    "nats-py",
])

# For running servers with pre-built wheels on localhost (not for docker or publishing)
ttt_requirements = merge_list(DEFAULT_REQUIREMENTS,parse_requirements("./src/gai/ttt/requirements_ttt.txt"))
ttt_requirements.append(local_pkg("exllamav2","exllamav2-0.1.4-cp310-cp310-linux_x86_64.whl"))

setup(
    name='gai-sdk',
    version=VERSION,
    author="kakkoii1337",
    author_email="kakkoii1337@gmail.com",
    package_dir={'': 'src'},
    packages=find_packages(where='src'),
    description = "Refer to https://gai-labs.github.io/gai for more information",
    long_description="Refer to https://gai-labs.github.io/gai for more information",
    long_description_content_type="text/markdown",
    classifiers=[
        'Programming Language :: Python :: 3.10',
        "Development Status :: 3 - Alpha",        
        'License :: OSI Approved :: MIT License',
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",        
        'Operating System :: OS Independent',
        "Topic :: Software Development",
        "Topic :: Software Development :: Libraries :: Python Modules",        
        "Topic :: Scientific/Engineering :: Artificial Intelligence",        
    ],
    python_requires='>=3.10',
    install_requires=[
        DEFAULT_REQUIREMENTS
    ],
    extras_require= {
        "cli": cli_requirements,
        "dev": dev_requirements,
        "ttt": ttt_requirements
    } if not is_publishing else {},
    entry_points={
        'console_scripts': [
            'gai=gai.cli.scripts.main:main',
        ],
    }
)


import os,torch,gc,json
from typing import Optional
from gai.lib.common.utils import get_app_path
from gai.lib.common.logging import getLogger
logger = getLogger(__name__)
from jsonschema import validate, ValidationError
from gai.lib.common.profile_function import profile_function
from gai.ttt.server.builders import CompletionsFactory

def correct_single_quote_json(s):
    rstr = ""
    escaped = False

    # Remove single quotes at the beginning and end of the string
    s=s.strip("'")

    for c in s:
    
        if c == "'" and not escaped:
            c = '"' # replace single with double quote
        
        elif c == "'" and escaped:
            rstr = rstr[:-1] # remove escape character before single quotes
        
        elif c == '"':
            c = '\\' + c # escape existing double quotes

        escaped = (c == "\\") # check for an escape character
        rstr += c # append the correct json
    
    return rstr

class GaiExLlamav2:

    def __init__(self, gai_config):
        if (gai_config is None):
            raise Exception("ExLlama_TTT2: gai_config is required")
        if gai_config.get("model_path",None) is None:
            raise Exception("ExLlama_TTT2: model_path is required")
        if gai_config.get("model_basename",None) is None:
            raise Exception("ExLlama_TTT2: model_basename is required")

        self.gai_config = gai_config
        self.model_dir = os.path.join(get_app_path(
        ), gai_config["model_path"])
        self.cache = None
        self.model = None
        self.tokenizer = None
        self.prompt = None
        self.generator = None
    
    def __enter__(self):
        self.load()
        return self
    
    def __exit__(self,exc_type, exc_value,traceback):
        self.unload()
        gc.collect()
        torch.cuda.empty_cache()


    @profile_function
    def load_config(self):
        from exllamav2 import ExLlamaV2Config
        config=ExLlamaV2Config()
        config.model_dir = self.model_dir
        config.prepare()
        config.max_seq_len = self.gai_config.get("max_seq_len",8192)
        config.no_flash_attn = self.gai_config.get("no_flash_attn",True)
        self.exllama_config = config

    @profile_function
    def load_model(self):
        from exllamav2 import ExLlamaV2
        self.model = ExLlamaV2(self.exllama_config)
    
    @profile_function    
    def load_cache(self):
        from exllamav2.cache import ExLlamaV2Cache_Q4
        self.cache = ExLlamaV2Cache_Q4(self.model, 
                                       lazy=True, 
                                       max_seq_len=self.gai_config.get("max_seq_len",8192))
        self.model.load_autosplit(self.cache)

    @profile_function
    def load_tokenizer(self):
        from exllamav2 import ExLlamaV2Tokenizer
        self.tokenizer = ExLlamaV2Tokenizer(self.exllama_config)

    @profile_function
    def load_generator(self):
        from exllamav2.generator import ExLlamaV2DynamicGenerator

        # Note: Dynamic Generator requires flash-attn 2.5.7+ to use paged attention and only supports Ampere GPUs or newer, otherwise set paged=False
        self.generator=ExLlamaV2DynamicGenerator(model=self.model, 
                                            cache=self.cache, 
                                            tokenizer=self.tokenizer,
                                            paged=False)
        self.generator.warmup()

    # initial load
    def load(self):
        self.load_config()
        self.load_model()
        self.load_cache()
        self.load_tokenizer()

    def unload(self):
        if self.generator is not None:
            del self.model
            del self.cache
            del self.tokenizer
            del self.generator
            import gc,torch
            gc.collect()
            torch.cuda.empty_cache()
        return self

    # Configure JSON Schema enforcement for tool call and response model
    def prepare_filters(self, tools, tool_choice, response_model):
        """
        1. tool_choice always takes precedence over schema.
           If schema is required, then tool_choice must be set to "none".
        2. If tool_choice is "auto", then tools schema will always be used but tools will include { "type": "text", "text": "..." }.
        3. If tool_choice is "required", then tools schema will always be used and tools will not include { "type": "text", "text": "..." }.
        4. If tool_choice is "none", then schema will be used if it is available.
        5. If tool_choice is "none" and schema is not available then output will always be text.
        6. If tool_choice is "none" and schema is available then output will always be based on schema, aka. JSON mode.        
        """

        self.validation_schema=None

        # Create filter from schema
        def create_filter(validation_schema,tokenizer):
            from lmformatenforcer.integrations.exllamav2 import build_token_enforcer_tokenizer_data
            tokenizer_data = build_token_enforcer_tokenizer_data(tokenizer)                
            from lmformatenforcer import JsonSchemaParser
            parser = JsonSchemaParser(validation_schema)
            from lmformatenforcer.integrations.exllamav2 import ExLlamaV2TokenEnforcerFilter
            from exllamav2.generator.filters.prefix import ExLlamaV2PrefixFilter
            return [ExLlamaV2TokenEnforcerFilter(parser, tokenizer_data),
                ExLlamaV2PrefixFilter(self.model, self.tokenizer, ["{","\n\n{"])]

        if not tools and tool_choice=="required":
            raise Exception("tool_choice='required' requires tools to be provided.")
        
        if (not self.is_validation_required(self.job_state)):
            self.validation_schema=None
            return None        

        if (self.is_using_tools(self.job_state)):
            # If toolcall is used, use toolcall schema and create filters for it
            from gai.lib.common.generators_utils import get_tools_schema
            self.validation_schema=get_tools_schema()
            return create_filter(validation_schema=self.validation_schema, tokenizer=self.tokenizer)        

        if (self.is_using_response_model(self.job_state)):
            # If user_defined schema, apply user_defined schema
            self.validation_schema=response_model
            return create_filter(validation_schema=self.validation_schema, tokenizer=self.tokenizer)

    # Configure intermediate prompts for tool call and response model formatted for the underlying LLM
    def prepare_prompt(self, messages, response_model, tools, tool_choice, stream):
        formatted_messages = messages

        if (self.is_using_tools(self.job_state)):

            # Add system_message for tool_call
            from gai.lib.common.generators_utils import apply_tools_message
            formatted_messages = apply_tools_message(messages=formatted_messages,
                tools=tools,
                tool_choice=tool_choice)
            
        if (self.is_using_response_model(self.job_state)):

            # Add system_message to use schema (response model) if tool_call are not applicable
            from gai.lib.common.generators_utils import apply_schema_prompt
            formatted_messages = apply_schema_prompt(messages=formatted_messages,
                schema=response_model)

        prompt_format = self.gai_config.get("prompt_format")
        from gai.lib.common.generators_utils import format_list_to_prompt
        return format_list_to_prompt(messages=formatted_messages, format_type=prompt_format, stream=stream)
    
    def is_validation_required(self,job_state):
        return self.is_using_tools(job_state) or self.is_using_response_model(job_state)

    # tools are only used when tool_choice is "required" and tools are available
    def is_using_tools(self,job_state):
        using_tools = (job_state["tool_choice"]=="required" and job_state["tools"])
        return using_tools

    # response model is only used when tool_choice is none and response_model is available
    def is_using_response_model(self,job_state):
        using_response_model = ((job_state["tool_choice"]=="none" or job_state["tool_choice"]=="auto") and job_state["response_model"])
        return using_response_model

    @profile_function
    def generate(self):

        eos = False
        while not eos:
            results = self.generator.iterate()
            for result in results:
                eos=result.get("eos",False)
        return result

    # Retry completions if validation is required
    def _generate_with_retries(self):
        retries=self.job_state["max_retries"]

        # Validation not required
        if not self.is_validation_required(self.job_state):
            return self.generate()
        
        # Validation required
        while retries>0:
            result=self.generate()
            try:
                if (self.is_using_tools(self.job_state)):
                    jsoned={
                        "type":"function",
                        "function":json.loads(result["full_completion"])
                    }
                    validate(instance=jsoned, schema=self.validation_schema)

                if (self.is_using_response_model(self.job_state)):
                    jsoned=json.loads(result["full_completion"])
                    validate(instance=jsoned, schema=self.validation_schema)

                # We are safe once we reach here since its either text or we have passed validations
                return result
            except ValidationError as e:
                logger.debug(f"GaiExLlamav2.generate_with_retries: Failed to validate {jsoned}.")
                retries-=1
                self.load_job()
            except Exception as e:
                logger.error(f"GaiExLlamav2.generate_with_retries: error={e}")
                raise e
        raise Exception("GaiExLlamav2.generate_with_retries: Validation of schema is required and failed after max retries.")

    def _streaming(self):
        eos = False
        result = None
        completed = ""

        if self.is_validation_required(self.job_state):
            raise Exception("GaiExLlamav2._streaming: Validation of schema is required and not supported in streaming mode.")

        while not eos:

            # Run one iteration of the generator. Returns a list of results
            results = self.generator.iterate()
            for result in results:
                text=result.get("text","")
                if text:
                    completed+=text
                    yield text
                eos=result.get("eos",False)

        yield result

    def load_job(self):

        from exllamav2.generator import ExLlamaV2Sampler
        settings = ExLlamaV2Sampler.Settings()
        settings.temperature=self.job_state["temperature"]
        settings.top_k=self.job_state["top_k"]
        settings.top_p=self.job_state["top_p"]
        settings = settings

        # Prepare settings.filters
        from exllamav2.generator import ExLlamaV2DynamicJob
        settings.filters = self.prepare_filters(tools=self.job_state["tools"], 
                                                     tool_choice=self.job_state["tool_choice"],
                                                     response_model=self.job_state["response_model"])
        if settings.filters:
            logger.info(f"GaiExLlamav2.load_job: apply filters.")
            settings.temperature=0
        prompt=self.prepare_prompt(messages=self.job_state["messages"],
                                   response_model=self.job_state["response_model"],
                                   tools=self.job_state["tools"],
                                   tool_choice=self.job_state["tool_choice"],
                                   stream=True)

        # Reset generator and run job
        self.job = ExLlamaV2DynamicJob(    
            input_ids = self.tokenizer.encode(prompt),
            gen_settings = settings,
            max_new_tokens = self.job_state["max_new_tokens"],
            completion_only = True,
            token_healing = True,
            seed = 1,
            stop_conditions=self.job_state["stop_conditions"],
            add_bos=False,
            #encode_special_tokens=self.job_state["encode_special_tokens"],
            decode_special_tokens=self.job_state["decode_special_tokens"],
        )
        # for pending in self.generator.pending_jobs:
        #     self.generator.cancel(pending)
        self.generator.enqueue(self.job)
        return self.generator               

    # Main driver for completions that calls _generate and _streaming,
    # but only return pure dictionary as output
    # Parameters are captured into a job_state for completions.
    def _create(self):
        try:
            self.load_job()
        except Exception as e:
            logger.error(f"GaiExLlamav2._create: Error loading job. error={e}")

        if not self.job_state["stream"]:
            return self._generate_with_retries()
        return (chunk for chunk in self._streaming())

    def initialize_job_state(self,
        messages:list,
        stream:Optional[bool], 
        tools:Optional[list]=None,
        tool_choice:Optional[str]=None,
        response_model=None,
        max_new_tokens:Optional[int]=None,
        stop_conditions:Optional[list]=None,
        temperature:Optional[float]=None,
        top_p:Optional[float]=None,
        top_k:Optional[int]=None,
        max_retries:int=None
        ):
        logger.info("gai_exllamav2.create: Initializing job_state.")
        if self.tokenizer.eos_token_id not in self.gai_config["stop_conditions"]:
            self.gai_config["stop_conditions"].append(self.tokenizer.eos_token_id)
        self.job_state = {
            "messages": messages,
            "stream": stream,
            "tools": tools,
            "response_model": response_model,
            "tool_choice": tool_choice or self.gai_config["tool_choice"],
            "stop_conditions": stop_conditions or self.gai_config["stop_conditions"],
            "prompt_format": self.gai_config["prompt_format"],
            "max_new_tokens": max_new_tokens or self.gai_config["hyperparameters"]["max_new_tokens"],
            "temperature": temperature or self.gai_config["hyperparameters"]["temperature"],
            "top_p": top_p or self.gai_config["hyperparameters"]["top_p"],
            "top_k": top_k or self.gai_config["hyperparameters"]["top_k"],  
            "max_retries": max_retries or self.gai_config["max_retries"],
            "decode_special_tokens": self.gai_config["decode_special_tokens"]
        }
        if self.is_validation_required(self.job_state):
            self.job_state["temperature"]=0        

    def create(self, 
        messages:list,
        stream:Optional[bool], 
        tools:Optional[list]=None,
        tool_choice:Optional[str]=None,
        response_model=None,
        max_new_tokens:Optional[int]=None,
        stop_conditions:Optional[list]=None,
        temperature:Optional[float]=None,
        top_p:Optional[float]=None,
        top_k:Optional[int]=None,
        max_retries:int=None
        ):
        self.load_generator()

        self.initialize_job_state(
            messages=messages,
            stream=stream,
            tools=tools,
            tool_choice=tool_choice,
            response_model=response_model,
            max_new_tokens=max_new_tokens,
            stop_conditions=stop_conditions,
            temperature=temperature,
            top_p=top_p,
            top_k=top_k,
            max_retries=max_retries
        )

        # Create completions
        response=self._create()

        # Convert Output
        factory = CompletionsFactory()        
        if not stream:
            if self.is_using_tools(self.job_state):
                logger.info(f"ExLlama_TTT2: factory.message.build_toolcall(response)")
                return factory.message.build_toolcall(response)
            else:
                logger.info(f"ExLlama_TTT2: factory.message.build_content(response)")
                return factory.message.build_content(response)
        else:
            return (chunk for chunk in factory.chunk.build_stream(response))



